// Copyright (c) 2002 Tom Barry.  All rights reserved.
//      trbarry@trbarry.com
// Requires Avisynth source code to compile for Avisynth
// Avisynth Copyright 2000 Ben Rudiak-Gould.
//      http://www.math.berkeley.edu/~benrg/avisynth.html
/////////////////////////////////////////////////////////////////////////////
//
//	This file is subject to the terms of the GNU General Public License as
//	published by the Free Software Foundation.  A copy of this license is
//	included with this software distribution in the file COPYING.  If you
//	do not have a copy, you may obtain a copy by writing to the Free
//	Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
//	This software is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details
//  
//  Also, this program is "Philanthropy-Ware".  That is, if you like it and 
//  feel the need to reward or inspire the author then please feel free (but
//  not obligated) to consider joining or donating to the Electronic Frontier
//  Foundation. This will help keep cyber space free of barbed wire and bullsh*t.  
//
/////////////////////////////////////////////////////////////////////////////
// Change Log
//
// Date          Developer             Changes
//
// 2004/08/19  v.0.1.0.3	Released by Fizick - http://bag.hotmail.ru, bag@hotmail.ru
//							Fixed bug with spatial filter radius for YV12
//
// 2004/08/10  V 0.1.0.2 	by Fizick - http://bag.hotmail.ru, bag@hotmail.ru - not public
//							Fix one main bug (for both YUY2 and YV12 modes):
//							in all previous versions light pixels were not filtered,
//							both S_EdgeThreshHold and T_MotionThreshHold did not work.
//
// 2003/06/22  V 0.1.0.1	Fix bug affecting YV12 only
//
// 2003/01/22  V 0.1.0.0	Remove chroma filtering. Add YV12 support. Use Pluginit2
//						  	Also remove Horizontal parms, add them to spatial.
//							Hopefully fix purple tint.

// 2002/08/03  V 0.0.3.0	Add some chroma filtering  (later removed, above)

// 2002/08/03  V 0.0.2.0	Bugs, Split out Spacial, Temporal, Horizontal user parms

// 2002/08/02  V 0.0.1.0	Initial test release 

// CVS Log
//
/////////////////////////////////////////////////////////////////////////////
//#include "internal.h"

#include "windows.h"
#include "avisynth.h"
#include <malloc.h>
#include <string.h>

#define BOOL unsigned int
#define BYTE unsigned char

const __int64 YMask   = 0x00ff00ff00ff00ff;
const __int64 UVMask  = 0xff00ff00ff00ff00;
const __int64 AllMask = 0xffffffffffffffff;

class STMedianFilter : public GenericVideoFilter {
    int  Max_Frames;
    __int64  SThreshHoldF;
    __int64  SThreshHoldE;
    __int64  TThreshHoldF;
    __int64  TThreshHoldM;
	int  useFrame;

    PVideoFrame src;
    PVideoFrame dst;
    PVideoFrame prevSrc;
    PVideoFrame nextSrc;
    unsigned char* dstp;


    int src_pitch;
    int dst_pitch;
    int rowsize;
    int height;
    int FldHeight;
	bool SSE2enabled;
	bool SSEMMXenabled;
	bool _3DNOWenabled;
	bool DoYV12;
	int __stdcall Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch);
	BOOL GoFilter_YUY2(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, const BYTE* pPrev, const BYTE* pNext,
			int FldHeight);				
	BOOL GoFilter_YV12(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, const BYTE* pPrev, const BYTE* pNext,
			int FldHeight);			

public:
    STMedianFilter(PClip _child, 
		int _SThreshHoldF, int _SThreshHoldE,
		int _TThreshHoldF, int _TThreshHoldM,
		 IScriptEnvironment* env) : GenericVideoFilter(_child)
    {
		_int64 i;
	
		i = _SThreshHoldF;	
        SThreshHoldF = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
		i = _SThreshHoldE;	
        SThreshHoldE = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
		i = _TThreshHoldF;	
        TThreshHoldF = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
		i = _TThreshHoldM;	
        TThreshHoldM = i << 56 | i << 48 | i << 40 | i << 32 | i << 24 | i << 16 | i << 8 | i; 
        Max_Frames = vi.num_frames;
		SSEMMXenabled = (env->GetCPUFlags() & CPUF_INTEGER_SSE) !=0;


	    if (vi.IsYUY2())
		{
			DoYV12 = false;
		}
		else if (vi.IsYV12())
		{
			DoYV12 = true;
		}
	    else
		{
			env->ThrowError("STMedianFilter: Supports YUY2 and YV12 color formats only");
		}

	    if (!SSEMMXenabled)
		{
			env->ThrowError("STMedianFilter: Only SSEMMX CPU's currently supported");
		}

    }

    PVideoFrame __stdcall GetFrame(int inFrame, IScriptEnvironment* env);

};


PVideoFrame __stdcall STMedianFilter::GetFrame(int inFrame, IScriptEnvironment* env) 
{
PVideoFrame PrevSrc;
PVideoFrame NextSrc;

const BYTE* srcp;
const BYTE* pPrev;
const BYTE* pNext;
	
	useFrame = inFrame;
	
	// Most of our processing is the same for YUY2 or YV12.
	src = child->GetFrame(useFrame, env);
	srcp = src->GetReadPtr(PLANAR_Y);

	if (useFrame < 1 || useFrame >= Max_Frames-1)
	{
		return src;						// don't filter first and last frames
	}

	PrevSrc = child->GetFrame(useFrame-1, env);
	pPrev = PrevSrc->GetReadPtr(PLANAR_Y);
	NextSrc = child->GetFrame(useFrame+1, env);
	pNext = NextSrc->GetReadPtr(PLANAR_Y);
	dst = env->NewVideoFrame(vi);
	dstp = dst->GetWritePtr(PLANAR_Y);
	src_pitch = src->GetPitch(PLANAR_Y);
	dst_pitch = dst->GetPitch(PLANAR_Y);
	rowsize = dst->GetRowSize(PLANAR_Y);
	height = dst->GetHeight();


	// But if it is YV12 we still have to copy the 2 chroma planes
	if (vi.IsYV12())
	{
		GoFilter_YV12(src_pitch, dst_pitch,
			rowsize, srcp, dstp, pPrev, pNext,
			height);					// go H expand lines into even output lines

		srcp = src->GetReadPtr(PLANAR_U);
		pPrev = PrevSrc->GetReadPtr(PLANAR_U);
		pNext = NextSrc->GetReadPtr(PLANAR_U);
		dstp = dst->GetWritePtr(PLANAR_U);
		src_pitch = src->GetPitch(PLANAR_U);
		dst_pitch = dst->GetPitch(PLANAR_U);
		rowsize = dst->GetRowSize(PLANAR_U);
		height = dst->GetHeight(PLANAR_U);
		env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
		
		srcp = src->GetReadPtr(PLANAR_V);
		pPrev = PrevSrc->GetReadPtr(PLANAR_V);
		pNext = NextSrc->GetReadPtr(PLANAR_V);
		dstp = dst->GetWritePtr(PLANAR_V);
		src_pitch = src->GetPitch(PLANAR_V);
		dst_pitch = dst->GetPitch(PLANAR_V);
		rowsize = dst->GetRowSize(PLANAR_V);
		height = dst->GetHeight(PLANAR_V);
		env->BitBlt(dstp, dst_pitch, srcp, src_pitch, rowsize, height);
	}
	else
	{
		GoFilter_YUY2(src_pitch, dst_pitch,
			rowsize, srcp, dstp, pPrev, pNext,
			height);					// go H expand lines into even output lines
	}

	__asm emms;

    return dst;
}

#define IS_SSE
#define SSE_TYPE SSE


#define ChkBetterNeighbors(PADDR1, PADDR2) \
__asm { \
		__asm movq	mm0, qword ptr[(PADDR1)] /* our 4 pixels */ \
		__asm movq mm1, qword ptr[(PADDR2)] /* our pixel2 value */ \
		__asm movq mm2, mm0			/* another copy of our pixel1 value */ \
		__asm pminub mm0, mm1       /* minimum value of the 2 */ \
		__asm pmaxub mm1, mm2       /* maximum value of the 2 */ \
		__asm movq mm2, mm1			/* another copy of our max value */ \
		__asm psubusb mm2, mm0 	    /* the diff */ \
		__asm movq mm3, mm2			/* another copy of our our weights (diff) */ \
		__asm psubusb mm3, mm5		/* nonzero where old weights lower, else 0 */ \
		__asm pcmpeqb mm3, mm4		/* now ff where new better, else 00	*/ \
		__asm pand	mm0, mm3		/* keep only better min pixels */ \
		__asm pand	mm1, mm3		/* keep only better max pixels */ \
		__asm pand  mm2, mm3		/* and weights */ \
		__asm pcmpeqb mm3, mm4		/* here ff where old better, else 00 */ \
		__asm pand	mm5, mm3		/* keep only better old weights */ \
		__asm pand  mm6, mm3        /* keep only better old min  */ \
		__asm pand  mm7, mm3        /* keep only better old max  */ \
		__asm por	mm5, mm2		/* merge new & old weights */ \
		__asm por	mm6, mm0        /* merge new & old min */\
		__asm por	mm7, mm1		/* merge new & old max */ \
}


// For YUY2 Support

BOOL	STMedianFilter::GoFilter_YUY2(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, const BYTE* _pPrev, const BYTE* _pNext,
			int FldHeight)					// go H expand lines into even output lines
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-10) >> 3);
__int64 SThreshHoldFW = SThreshHoldF;
__int64 SThreshHoldEW = SThreshHoldE;
__int64 TThreshHoldFW = TThreshHoldF;
__int64 TThreshHoldMW = TThreshHoldM;
__int64 SaveTemp;
__int64	ChromaMask	  = (DoYV12)
						? AllMask
						: YMask;

	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pit*(FldHeight-1),
		pSrc+src_pit*(FldHeight-1), rowsize);  // and last

	pDest += dst_pit;


	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pit 
		mov		edx, pDest 

		// do first and last qwords, just move
		movq	mm0, qword ptr[ebx+ecx]		// just move 1st 4 pixels
		movq	qword ptr[edx], mm0
		
		mov		eax, rowsize
		add		ebx, eax
		movq	mm0, qword ptr[ebx+ecx-8]		// just move last 4 pixels
		movq	qword ptr[edx+eax-8], mm0

		sub		ebx, eax						// restore
		pxor	mm4, mm4						// useful constant 000..	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

// Search between prev and next frame - Temporal
        ChkBetterNeighbors(edi-2, esi+2*ecx+2)  // up left, down right
        ChkBetterNeighbors(edi+2, esi+2*ecx-2)  // up right, down left
        ChkBetterNeighbors(edi+2*ecx-2, esi+2)  // down left, up right
        ChkBetterNeighbors(edi+2*ecx+2, esi-2)   // down right, up left   
        ChkBetterNeighbors(edi+2*ecx, esi)		// down, up
        ChkBetterNeighbors(edi, esi+2*ecx)		// up, down
        ChkBetterNeighbors(edi+ecx-2, esi+ecx+2) // left, right
        ChkBetterNeighbors(edi+ecx+2, esi+ecx-2) // right, left
		ChkBetterNeighbors(edi+ecx, esi+ecx)	// center, in old and new

// We now know the min and max of the closest together pair of neighbors
// We will clip our current 4 pixels to (each) be withing this range if it would
// not cause them to change by more than the ThreshHold user parm
		movq	mm0, qword ptr[ebx+ecx]		// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor
		movq	mm3, mm1
		movq	mm2, mm0
		psubusb	mm3, mm0					// changed - orig
		psubusb mm2, mm1					// orig - changed
		por		mm3, mm2					// the abs diff
		psubusb	mm3, TThreshHoldFW			// did it change too much, Y where nonzero
		psubusb mm5, TThreshHoldMW			// Too much motion? Y where nonsero
		por		mm3, mm5 //mm2	// Don't clip where either nonzero - Bug (mm2 was in place of mm5) fixed by Fizick in v.0.1.0.2 
		pcmpeqb mm3, mm4					// now ff where new value should be used, else 00
		pand	mm3, ChromaMask				// but maybe only filter Luma
		pand	mm1, mm3					// use new value for these pixels
		pandn	mm3, mm0					// use old value for these pixels
		por		mm1, mm3					// combine them
		movq	SaveTemp, mm1				// save results for later

		// Do again to search spacial (both pixels in curr frame)
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

        ChkBetterNeighbors(ebx-2, ebx+2*ecx+2)  // up left, down right
        ChkBetterNeighbors(ebx+2, ebx+2*ecx-2)  // up right, down left
        ChkBetterNeighbors(ebx, ebx+2*ecx)	// up, down
        ChkBetterNeighbors(ebx+ecx-2, ebx+ecx+2) // left, right 

		movq	mm0, SaveTemp				// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor

		movq	mm3, mm1
		movq	mm2, mm0
		psubusb	mm3, mm0					// changed - orig
		psubusb mm2, mm1					// orig - changed
		por		mm3, mm2					// the abs diff
		psubusb	mm3, SThreshHoldFW			// did it change too much, Y where nonzero
		psubusb mm5, SThreshHoldEW			// Too much motion? Y where nonsero
		por		mm3, mm5 //mm2	// Don't clip where either nonzero - Bug (mm2 was in place of mm5) fixed by Fizick in v.0.1.0.2 
		pcmpeqb mm3, mm4					// now ff where new value should be used, else 00	
		pand	mm3, ChromaMask				// but maybe only filter Luma

		pand	mm1, mm3					// use new value for these pixels
		pandn	mm3, mm0					// use old value for these pixels
		por		mm1, mm3					// combine them, leave answer sitting in mm0

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ
		}

	// adjust for next line
	pSrc  += src_pit;
	pPrev  += src_pit;
	pNext  += src_pit;
	pDest += dst_pit;
	}
	return 0;
}	

//--------------------- YV12 -----------------------------------------------
// Same function as above except for offset to adjacent pixels a and chroma mask
	BOOL	STMedianFilter::GoFilter_YV12(int src_pit, int dst_pit,
			int rowsize, const BYTE* srcp, BYTE* dstp, const BYTE* _pPrev, const BYTE* _pNext,
			int FldHeight)					// go H expand lines into even output lines
{

const BYTE* pPrev = _pPrev;
const BYTE* pNext = _pNext;
const BYTE* pSrc = srcp;

BYTE* pDest = dstp;
int y;
int ct = ((rowsize-10) >> 3);
__int64 SThreshHoldFW = SThreshHoldF;
__int64 SThreshHoldEW = SThreshHoldE;
__int64 TThreshHoldFW = TThreshHoldF;
__int64 TThreshHoldMW = TThreshHoldM;
__int64 SaveTemp;
//__int64	ChromaMask	  = (DoYV12)
//						? AllMask
//						: YMask;

	memcpy(pDest, pSrc, rowsize);		// copy first line
	memcpy(pDest+dst_pit*(FldHeight-1),
		pSrc+src_pit*(FldHeight-1), rowsize);  // and last

	pDest += dst_pit;


	for (y=1; y <= FldHeight-2; y++)	
	{

		__asm
		{
		// Loop general reg usage
		//
		// eax - inner loop counter
		// ebx - src lines above curr 
		// ecx - src pitch
		// edx - dest
		// edi - src pixels, next frame, 1 line up
		// esi - src pixels, prev frame, 1 line up

		mov		ebx, pSrc   
		mov		esi, pPrev
		mov		edi, pNext
		mov		ecx, src_pit 
		mov		edx, pDest 

		// do first and last qwords, just move
		movq	mm0, qword ptr[ebx+ecx]		// just move 1st 4 pixels
		movq	qword ptr[edx], mm0
		
		mov		eax, rowsize
		add		ebx, eax
		movq	mm0, qword ptr[ebx+ecx-8]		// just move last 4 pixels
		movq	qword ptr[edx+eax-8], mm0

		sub		ebx, eax						// restore
		pxor	mm4, mm4						// useful constant 000..	
		// make some adjustments to start output on 2nd qword of 2nd line
		add		edi, 8
		add		ebx, 8
		add		esi, 8
		add		edx, 8
		mov	    eax, ct			// loop ct

LoopQ:	
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

// Search between prev and next frame - Temporal
        ChkBetterNeighbors(edi-1, esi+2*ecx+1)  // up left, down right
        ChkBetterNeighbors(edi+1, esi+2*ecx-1)  // up right, down left
        ChkBetterNeighbors(edi+2*ecx-1, esi+1)  // down left, up right
        ChkBetterNeighbors(edi+2*ecx+1, esi-1)   // down right, up left   
        ChkBetterNeighbors(edi+2*ecx, esi)		// down, up
        ChkBetterNeighbors(edi, esi+2*ecx)		// up, down
        ChkBetterNeighbors(edi+ecx-1, esi+ecx+1) // left, right
        ChkBetterNeighbors(edi+ecx+1, esi+ecx-1) // right, left
		ChkBetterNeighbors(edi+ecx, esi+ecx)	// center, in old and new

// We now know the min and max of the closest together pair of neighbors
// We will clip our current 4 pixels to (each) be withing this range if it would
// not cause them to change by more than the ThreshHold user parm
		movq	mm0, qword ptr[ebx+ecx]		// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor
		movq	mm3, mm1
		movq	mm2, mm0
		psubusb	mm3, mm0					// changed - orig
		psubusb mm2, mm1					// orig - changed
		por		mm3, mm2					// the abs diff
		psubusb	mm3, TThreshHoldFW			// did it change too much, Y where nonzero
		psubusb mm5, TThreshHoldMW			// Too much motion? Y where nonsero
		por		mm3, mm5 //mm2	// Don't clip where either nonzero - Bug (mm2 was in place of mm5) fixed by Fizick in v.0.1.0.2 
		pcmpeqb mm3, mm4					// now ff where new value should be used, else 00
//		pand	mm3, ChromaMask				// but maybe only filter Luma
		pand	mm1, mm3					// use new value for these pixels
		pandn	mm3, mm0					// use old value for these pixels
		por		mm1, mm3					// combine them
		movq	SaveTemp, mm1				// save results for later

		// Do again to search spacial (both pixels in curr frame)
		pcmpeqb	mm5, mm5		// init diff 255's
		pxor	mm6, mm6		// init min to 0
		pcmpeqb	mm7, mm7		// init max 255's

        ChkBetterNeighbors(ebx-1, ebx+2*ecx+1)  // up left, down right // bug fixed in v.0.1.0.3 by Fizick
        ChkBetterNeighbors(ebx+1, ebx+2*ecx-1)  // up right, down left // bug fixed in v.0.1.0.3 by Fizick
        ChkBetterNeighbors(ebx, ebx+2*ecx)	// up, down
        ChkBetterNeighbors(ebx+ecx-1, ebx+ecx+1) // left, right // bug fixed in v.0.1.0.3 by Fizick

		movq	mm0, SaveTemp				// our curr pixels vals
		movq	mm1, mm0
		pminub	mm1, mm7					// make it no larger than max neighbor
		pmaxub	mm1, mm6					// and now smaller than min neighbor

		movq	mm3, mm1
		movq	mm2, mm0
		psubusb	mm3, mm0					// changed - orig
		psubusb mm2, mm1					// orig - changed
		por		mm3, mm2					// the abs diff
		psubusb	mm3, SThreshHoldFW			// did it change too much, Y where nonzero
		psubusb mm5, SThreshHoldEW			// Too much motion? Y where nonsero
		por		mm3, mm5 //mm2	// Don't clip where either nonzero - Bug (mm2 was in place of mm5) fixed by Fizick in v.0.1.0.2 
		pcmpeqb mm3, mm4					// now ff where new value should be used, else 00	
//		pand	mm3, ChromaMask				// but maybe only filter Luma

		pand	mm1, mm3					// use new value for these pixels
		pandn	mm3, mm0					// use old value for these pixels
		por		mm1, mm3					// combine them, leave answer sitting in mm0

		movntq	qword ptr[edx], mm1			// and store it
		add		esi, 8
		add		edi, 8
		add		ebx, 8
		add		edx, 8

		dec		eax
		jnz		LoopQ
		}

	// adjust for next line
	pSrc  += src_pit;
	pPrev  += src_pit;
	pNext  += src_pit;
	pDest += dst_pit;
	}
	return 0;
}

int STMedianFilter::Fieldcopy(void *dest, const void *src, size_t count, 
				int rows, int dst_pitch, int src_pitch)
{
BYTE* pDest = (BYTE*) dest;
BYTE* pSrc = (BYTE*) src;
	
	for (int i=0; i < rows; i++)
	{
		memcpy(pDest, pSrc, count);
		pSrc += src_pitch;
		pDest += dst_pitch;
	}
	return 0;
}


AVSValue __cdecl Create_STMedianFilter(AVSValue args, void* user_data,
IScriptEnvironment* env) 
{
    return new STMedianFilter(args[0].AsClip(), args[1].AsInt(), args[2].AsInt(),
		args[3].AsInt(), args[4].AsInt(), env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("STMedianFilter", "ciiii", Create_STMedianFilter, 0);
    return "`STMedianFilter' Clip in Space and Time";
}

